//============================================================================
//  Copyright (c) 2000 University of Kansas 
//  All rights reserved.
//
//  KU MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF 
//  THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
//  TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
//  PARTICULAR PURPOSE, OR NON-INFRINGEMENT.  KU SHALL NOT BE LIABLE
//  FOR ANY DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING,
//  MODIFYING OR DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
//
//
//  By using or copying this Software, Licensee agrees to abide by the
//  intellectual property laws, and all other applicable laws of the
//  U.S., and the terms of this license.
//
//	authors: Cindy ckong@ittc.ukans.edu
//
//  $Id: SOM_RecordType.java,v 1.2 2001/04/04 22:23:20 cvs Exp $
//
//  Revision History:
//  ----------------
//  $Log: SOM_RecordType.java,v $
//  Revision 1.2  2001/04/04 22:23:20  cvs
//  Removed an error in typechecking of record. Type checking now works
//  for recursive functions.
//  -- Cindy
//
//  Revision 1.1  2000/07/24 23:41:02  cvs
//  Some type checking added for record and sequence.
//  -- Cindy
//
//
//=============================================================================
package SOM;

import util.ListEnumeration;
import java.util.Hashtable;
import java.util.Enumeration;
import SYM.SYM_TypeInconsistencyException;

/** This represents the <B>RecordType</B> object
 * @author <I>Cindy Kong</I> (ckong@ittc.ukans.edu)
 */
public class SOM_RecordType
extends SOM_TypeDeclaration {

	// The fields in the record
	// The id of the field is the key and the value associated with
	// the key, the type.
	Hashtable fields = new Hashtable();

	/** The primary constructor
	 * @param myType The SOM_TypeDeclaration corresponding to the
	 * "bunch MetaType"
	 * @param myId The Identifier containing the name of the type
	 * @return Not applicable
	 */
	public SOM_RecordType(SOM_Identifier myId, SOM_TypeDeclaration myType) {
		super(myId, myType);
	}

	/** Another constructor
	 * @param myType The SOM_TypeDeclaration corresponding to the
	 * "bunch MetaType"
	 * @return Not applicable
	 */
	public SOM_RecordType(SOM_TypeDeclaration myType) {
		// The default name is setType
		this(new SOM_Identifier("recordType"), myType);
	}

	/** To set the members and their type
	 * @param idIn The SOM_Identifier for the name of the field
	 * @param memberTypeIn The SOM_TypeDeclaration corresponding to field
	 * @return void
	 */
	public void setFieldType(SOM_Identifier idIn, 
							 SOM_TypeDeclaration memberTypeIn) {
		fields.put(idIn, memberTypeIn);
	}

	/** To get the member type associated with a field
	 * @param idIn The SOM_Identifier for the name of the field
	 * @return The SOM_TypeDeclaration of member
	 */
	public SOM_TypeDeclaration getMemberType(SOM_Identifier idIn) {
		return (SOM_TypeDeclaration)fields.get(idIn);
	}

	/** To get the fields
	 * @return The Hashtable containing the fields
	 */
	public Hashtable getFields() {
		return fields;
	}

	/** To implement a version of isDistinguishable for record types
	 * @param typeOne The first SOM_RecordType
	 * @param typeTwo The next SOM_RecordType to be distinguished
	 * against typeOne
	 * @return true if they are distinguishable, false otherwise
	 */
	public static boolean isDistinguishable(SOM_RecordType typeOne,
											SOM_RecordType typeTwo) { 
		Hashtable fieldsOne = typeOne.getFields();
		Hashtable fieldsTwo = typeTwo.getFields();
		Enumeration keysOne = fieldsOne.keys();
		Object key;
		SOM_TypeDeclaration fieldTypeOne;
		SOM_TypeDeclaration fieldTypeTwo;

		if (fieldsOne.equals(fieldsTwo))
			return false;

		if (fieldsOne.size() != fieldsTwo.size())
			return true;
			
		while (keysOne.hasMoreElements()) {
			key = keysOne.nextElement();
			fieldTypeOne = (SOM_TypeDeclaration)fieldsOne.get(key);
			fieldTypeTwo = (SOM_TypeDeclaration)fieldsTwo.get(key);
			if (SOM_TypeDeclaration.isDistinguishable
				(fieldTypeOne, fieldTypeTwo))
				return true;
		}
		
		return false;
	}

	/** To ascertain "compatiblity" for bindings of the kind lhs <- rhs
	 * @param typeLHS The RecordType corresponding to the lhs object
	 * @param typeRHS The RecordType corresponding to the rhs object
	 * @return true, if the rhs is assignable to the lhs, false otherwise
	 */
	public static boolean
			isLeftTypeCompatibleWithRightType(SOM_RecordType typeLHS,
											  SOM_RecordType typeRHS) { 
		Hashtable fieldsLHS = typeLHS.getFields();
		Hashtable fieldsRHS = typeRHS.getFields();
		Enumeration keysLHS = fieldsLHS.keys();
		Enumeration keysRHS;
		Object keyLHS;
		Object keyRHS = null;		
		SOM_TypeDeclaration fieldTypeLHS;
		SOM_TypeDeclaration fieldTypeRHS;
		String keyId;

		if (fieldsRHS.equals(fieldsLHS))
			return true;

		if (fieldsRHS.size() != fieldsLHS.size())
			return false;
			
		while (keysLHS.hasMoreElements()) {
			keyLHS = keysLHS.nextElement();
			keyId = keyLHS.toString();
			
			keysRHS = fieldsRHS.keys();
			while (keysRHS.hasMoreElements()) {
				keyRHS = keysRHS.nextElement();
				if (keyRHS.toString().equals(keyId)) {
					break;
				}
			}
			if (!keyRHS.toString().equals(keyId)) {
				throw new SYM.SYM_TypeInconsistencyException
					("Field with key: " + keyId + " missing in RHS");
			}
			fieldTypeRHS = (SOM_TypeDeclaration)fieldsRHS.get(keyRHS);
			fieldTypeLHS = (SOM_TypeDeclaration)fieldsLHS.get(keyLHS);
			if (fieldTypeLHS == null) { 
				throw new NullPointerException("No type corresponding" +
											   " to field name: " + 
											   keyId);
			}
			if (fieldTypeRHS == null) { 
				throw new NullPointerException("No type corresponding" +
											   " to field name: " + 
											   keyId);
			}
			if (!fieldTypeRHS.isPredicateSubtypeOf(fieldTypeRHS))
				return false;
		}
		
		return true;
	}
}


