#!/usr/local/bin/perl

#-----------------------------------------------------------------------------
# Copyright (c) 2000 University of Kansas.
# All rights reserved.

# KU MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF 
# THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
# TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
# PARTICULAR PURPOSE, OR NON-INFRINGEMENT.  KU SHALL NOT BE LIABLE
# FOR ANY DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING,
# RESULT OF USING, MODIFYING OR DISTRIBUTING THIS SOFTWARE OR ITS
# DERIVATIVES.

# By using or copying this Software, Licensee agrees to abide by the
# intellectual property laws, and all other applicable laws of the
# U.S., and the terms of this license.
#-----------------------------------------------------------------------------

#
# Author: Darryl Dieckman, ddieckma@ececs.uc.edu


#
#
# This script will run regression tests on the VSPEC parser.
# The script will run clean tests on all the test cases located in the 
# clean subdirectory, and dirty tests on all the test cases located in
# the dirty subdirectory.
#
# Clean Tests:
# Clean test cases must pass the following tests:
#    Parse of test case
#    * The parser must return a zero exit status after parsing the test case
#    * The output of the parser must match the expected output
#
#    Generation of VSPEC
#    * The parser must return a zero exit status when printing VSPEC 
#    * The output of the parser must match the expected output
#    * The generated VSPEC must match the expected VSPEC
#    * The parser must return a zero exit status when re-parsing
#      the generated VSPEC.
#
#    Generation of LSL
#    * The parser must return a zero exit status when generating LSL traits
#    * The output of the parser must match the expected output
#    * The generated LSL traits must match the expected LSL traits
#    * The generated LSL traits must parse without error
#
#  Dirty Tests:
#    Parse of test case
#    * The parser must return a non-zero exit status when parsing a dirty
#      test case.
#    * The output of the parser must match the expected output
#    


use Cwd;
use File::Copy;

sub LogMsg {
    for($level = 0;$level < $logLevel; $level++){
	print $logfile "  ";
    }
    print $logfile @_[0];
}
sub PushLog {
    $logLevel++;
}
sub PopLog {
    $logLevel--;
}
$testLogCount = 1;
@testLogCase;
@testLogExpl;
@testLogFiles;
@testLogStatus;
sub TestResults {
    # Test Case
    # Explanation of test
    # Status (PASS/FAIL)
    # Results - reason for failure
    # Relevant files
    $testLogCase[$testLogCount] = shift @_;
    $testLogExpl[$testLogCount] = shift @_;
    $testLogStatus[$testLogCount] = shift @_;
    $testLogResults[$testLogCount] = shift @_;
    $testLogFiles[$testLogCount] = shift @_;
    $testLogCount++;
}
sub SummarizeResults {
    open(testResults,">test.results");
    print testResults 
	"==========================================================================\n";
    print testResults 
	"                                  SUMMARY\n";
    print testResults 
	"==========================================================================\n";
    
    for($i = 1; $i < $testLogCount; $i++){
	printf testResults "%3d. %-30s %-30s %s\n"
	    ,$i,$testLogCase[$i],$testLogExpl[$i], $testLogStatus[$i];
	
	if($testLogStatus[$i] eq $failed){
	    #
	    # Print explanation of error
	    #
	    print testResults "       Explanation:\n";
	    for $array_ref (@testLogResults[$i]){
		foreach $text (@$array_ref){
		    print testResults "           $text\n";
		}
	    }
	    #
	    # Print list of relevant files
	    #
	    print testResults "       See Files:\n";
	    for $array_ref (@testLogFiles[$i]){
		foreach $file (@$array_ref){
		    print testResults "           $file\n";
		}
	    }
	}
    }
    close(testResults);
    #system("less test.results");
    LogMsg("Summary of test results stored in 'test.results'\n");
}

sub CompareFiles {
    local($expectFileName) = shift @_;
    local($resultFileName) = shift @_;
    local($status) = 0;
    local($lineNum) = 0;
    
    # Check that files exist
    if(! -e $expectFileName){
	return -1;
    }
    if(! -e $resultFileName){
	return -2;
    }
    
    open(resultFile,$resultFileName);
    open(expectFile,$expectFileName);
    do {
	$expectLine = <expectFile>;
	$resultLine = <resultFile>;
	$lineNum++;
	if($expectLine =~ /VSPEC Parser, ver(.*)/ &&
	   $resultLine =~ /VSPEC Parser, ver(.*)/){
	    # Ignore the version  number
	} elsif($expectLine =~ /Copyright\(C\)(.*)/ &&
		$resultLine =~ /Copyright\(C\)(.*)/){
	    # Ignore the copyright
	} elsif($expectLine ne $resultLine){
	  print "Expected line:\n  $expectLine \n resultline \n $resultLine \n";
	   
	    if($status == 0){
		# Dump a header 
		LogMsg("\n");
		LogMsg("==> Differences between expected and actual results\n");
		LogMsg("==> expect file: $expectFile\n");
		LogMsg("==> result file: $resultFile\n");
	    }
	    # Print the difference
	    LogMsg("  expect($lineNum): $expectLine");
	    LogMsg("  result($lineNum): $resultLine");
	    $status = 1;
	    
	}
	
    } while(!eof(expectFile));
    
    if(eof(resultFile) == 0){
	LogMsg("==> Expected results file shorter than actusl\n");
	LogMsg("==> expect file: $expectFile\n");
	LogMsg("==> result file: $resultFile\n");
	while(!eof(resultFile)){
	    $resultLine = <resultFile>;
	    $lineNum++;
	    LogMsg("  result($lineNum): $resultLine");
	}
	return 1;
    }
    
    return $status;
}

sub DiffExpected {
    local($resultFile) = shift @_;
    local($expectFile) = shift @_;
    
    
    local($status) = CompareFiles($expectFile, $resultFile);
    
    
    if($status == 1){
	# There are differences between the expected and result files
	if ($updateExpected eq "force"){
		print "\n";
		print "forcing the expected output file to copy data
from result file";
		 LogMsg("Updating expected file $expectFile from
$resultFile\n");
		copy ($resultFile, $expectFile);
		return 1;
		}elsif ($updateExpected eq "query"){
		print "\n";
		print "DO YOU WANT TO UPDATE THE EXPECT FILE $expectFile 
y/n ";
		if (<> =~ "y"){
		print "coping $expectFile from $resultFile";
 LogMsg("Updating expected file $expectFile from $resultFile\n");
		copy ($resultFile, $expectFile);
		return 0;
		}
		LogMsg ("there are differences between the expected output file and the result file ");
		return 2;
		}else {
		LogMsg ("there are differences between the expected output file and the result file "); 
		return 2;
	    }
	
    } elsif ($status == -1){
	if ($updateExpected eq "force"){
		print "\n";
                print "forcing the expected output file to copy data
from result file";
                 LogMsg("Updating expected file $expectFile from
$resultFile\n");
                copy ($resultFile, $expectFile);
                return 1;
                }elsif ($updateExpected eq "do"){
		print "\n";
		print "coping the $resultFile to the $expectFile";
		 LogMsg("Updating expected file $expectFile from
$resultFile\n since this test file is new");
		copy ($resultFile, $expectFile);
		return 3;
		} else
		{
		  LogMsg ("The expected file does not exist");
		return 4;
		}
    } elsif ($status == -2){
	LogMsg("ERROR! The result '$resultFile' does not exist. Please investigate.");
	return 5;
    } else {
	return 0;
    }
    
    
}

sub DirtyTests {
    # Process arguments
    local($dirtyDir) = shift @_;
    
    #
    # Get list of test files
    #
    opendir(DIR, $dirtyDir);
    my(@files) = readdir(DIR);
    closedir(DIR);
    
    # Go to test director
    chdir "$dirtyDir" ||
	die "Can't locate dirty test directory $dirtyDir";
    
    # Setup results directory
    if(! (-d "results")){
	LogMsg("Creating results directory\n");
	mkdir("results",0777) ||
	    die "Unable to create results directory";
    }
    
    # Process test files
    $dir = cwd();
    foreach $file (@files) {
	if($file =~ /\S+\.sld$/i){
	    LogMsg("Testing file: $file\n");
	    #
	    # Run parse tests
	    #
	    ParseDirtyTest($file);
	}
    }
}

sub RemoveResultFile {
    local($resultFile) = shift @_;
    
    if(-e $resultFile){
	unlink($resultFile) ||
	    die "Unable to remove result file $resultFile";
    }
    
}

sub ParseDirtyTest {
    PushLog();
    # Process arguments
    local($fileBase) = shift @_;
    local($passTest) = 1;
    local($parseError) = 0;
    
    local($testName) = "Dirty Parse test";
    LogMsg("$testName...");
    
    
    local($resultFile) = "$resultDir$fileBase.parsetest";
    local($expectFile) = "$expectDir$fileBase.parsetest";
    RemoveResultFile($resultFile);
    
    local($status) = system("$vsp_cmd $file 2> $resultFile 1>> /dev/null");
    
    if($status == 0){
	LogMsg("$failed\n");
	TestResults($fileBase, $testName, $failed, 
		    ["The parser accepted a file that should contain and error"],
		    [$resultFile]);
	$passTest = 0;
    } else {
	LogMsg("$passed\n");
	TestResults($fileBase, $testName, $passed, [], []);
    }
    
    $testName = "Expected parse output";
    LogMsg("$testName...");
    $status = DiffExpected($resultFile, $expectFile);
    if($status == 0){
	LogMsg("$passed\n");
	TestResults($fileBase, $testName, $passed, [], []);
      } elsif ($status == 1){
	LogMsg("$passed \n");
	TestResults($fileBase, $testName, $passed, 
		    ["forcefully coping the expected output file from the results"],
		    [$resultFile ,$expectFile]);
    } elsif ($status == 2) {
	LogMsg("$failed\n");
	$passTest = 0;
	TestResults($fileBase, $testName, $failed, 
		    ["There are differences between the expected parser",
		     "output and the output of the test."],
		    [$resultFile, $expectFile]);
      } elsif ($status == 3){
	LogMsg("$passed\n");
    TestResults($fileBase, $testName, $passed,
		    ["the test file is new and so the expected output file is copied from the result file"],
		   [$expectFile]);
      }elsif ($status == 4){
	LogMsg("$failed \n");
	$passTest =0;
TestResults($fileBase, $testName, $failed,
		    ["The expected file doesnot exist !please investigate"],
		   [$expectFile]);	
      }elsif ($status == 5){
		LogMsg("$failed \n");
		$passTest =0;
	TestResults($fileBase, $testName, $failed,
		    ["The result  file doesnot exist !please investigate"],
		   [$resultFile]);
	      }

    PopLog();
    
    return $passTest;
}


sub ParseTest {
    PushLog();
    # Process arguments
    local($fileBase) = shift @_;
    local($passTest) = 1;
    local($parseError) = 0;
    
    local($testName) = "Parse test";
    LogMsg("$testName...");
    
    
    local($resultFile) = "$resultDir$fileBase.parsetest";
    local($expectFile) = "$expectDir$fileBase.parsetest";
    RemoveResultFile($resultFile);
    local($status) = system("$vsp_cmd $file 2> $resultFile 1>> /dev/null");
    
    if($status == 0){
	LogMsg("$passed\n");
	TestResults($fileBase, $testName, $passed, [], []);
    } else {
	LogMsg("$failed\n");
	TestResults($fileBase, $testName, $failed, 
		    ["The parser returned an error when parsing this SLDL file"],
		    [$resultFile]);
	$passTest = 0;
	$parseError = 1;
    }
    
    $testName = "Expected parse output";
    LogMsg("$testName...");
    $status = DiffExpected($resultFile, $expectFile);
    if($status == 0){
	LogMsg("$passed\n");
	TestResults($fileBase, $testName, $passed, [], []);
    } elsif ($status == 1){
	LogMsg("$passed \n");
	TestResults($fileBase, $testName, $passed, 
		    ["forcefully coping the expected output file from the results"],
		    [$resultFile ,$expectFile]);
    } elsif ($status == 2) {
	LogMsg("$failed\n");
	$passTest = 0;
	TestResults($fileBase, $testName, $failed, 
		    ["There are differences between the expected parser",
		     "output and the output of the test."],
		    [$resultFile, $expectFile]);
      } elsif ($status == 3){
	LogMsg("$passed\n");
    TestResults($fileBase, $testName, $passed,
		    ["the test file is new and so the expected output file is copied from the result file"],
		   [$expectFile]);
      }elsif ($status == 4){
	LogMsg("$failed \n");
	$passTest =0;
TestResults($fileBase, $testName, $failed,
		    ["The expected file doesnot exist !please investigate"],
		   [$expectFile]);	
      }elsif ($status == 5){
		LogMsg("$failed \n");
		$passTest =0;
	TestResults($fileBase, $testName, $failed,
		    ["The result  file doesnot exist !please investigate"],
		   [$resultFile]);
	      }
    
    PopLog();
    
    return $parseError;
}

sub LSLTest {
    PushLog();
    # Process arguments
    local($fileBase) = shift @_;
    local($passTest) = 1;
    
    local($testName) = "Publish LSL";
    LogMsg("$testName...");
    local($resultFile) = "$resultDir$fileBase.lsltest";
    local($expectFile) = "$expectDir$fileBase.lsltest";
    RemoveResultFile($resultFile);
    
    #
    # Generate LSL traits for the test file
    #
    local($status) = 
	system("$vsp_cmd -lsl $file 1> /dev/null 2> $resultFile");

    #
    # Check status of the VSPEC parser
    #
    if($status == 0){
	LogMsg("$passed\n");
	TestResults($fileBase, $testName, $passed, [], []);
    } else {
	LogMsg("$failed\n");
	$passTest = 0;
	TestResults($fileBase, $testName, $failed, 
		    ["The parser returned an error while publishing LSL"],
		    [$resultFile]);
    }
    
    #
    # Compare expected VSPEC parser output with actual VSPEC parser output
    #
    $testName = "Expected publish output";
    LogMsg("$testName...");
    $status = DiffExpected($resultFile, $expectFile);
    if($status == 0){
	LogMsg("$passed\n");
	TestResults($fileBase, $testName, $passed, [], []);
    } elsif ($status == -1){
	$passTest = 0;
	LogMsg("$failed\n");
	TestResults($fileBase, $testName, $failed, 
		    ["Unable to locate the expected output file!"],
		    [$expectFile]);
    } else {
	LogMsg("$failed\n");
	$passTest = 0;
	TestResults($fileBase, $testName, $failed, 
		    ["There are differences between the expected parser",
		     "output and the output of the test."],
		    [$resultFile, $expectFile]);
    }

    #
    # Collect the names of the generated traits
    #
    open(traitFile,"$resultFile");
    local(@traits);
    local(@traitResults);
    local($traitCount) = 0;
    while(<traitFile>){
	if(/stored in (.*\.lsl)/){
	    $traits[$traitCount] = $1;
	    $traitResults[$traitCount] = "$fileBase.$1";
	    #
	    # Copy trait to results directory
	    #
	    local($resultTrait) = "$resultDir$traitResults[$traitCount]";
	    copy($traits[$traitCount], $resultTrait);
	    $traitCount++;
	}
    }
    
    #
    # Parse generated LSL traits
    #
    $testName = "Parse LSL";
    LogMsg("$testName...\n");
    PushLog();
    foreach $trait (@traits){
	LogMsg("Parse trait $trait...");
	local($status) = system("$lsl_cmd $trait 1> /dev/null 2> /dev/null");
	if($status == 0){
	    LogMsg("$passed\n");
	    TestResults($fileBase, $testName, $passed, [], []);
	    unlink($trait);
	} else {
	    LogMsg("$failed\n");
	    $passTest = 0;
	    TestResults($fileBase, $testName, $failed, 
			["The LSL checker returned an error code while parsing",
			 "the generated trait.  See trait in the results directory"],
			[$trait]);
	}
	#
	# Remove the generated trait file, there is still a copy in the 
	# results directory.
	#
	unlink($trait);
    }
    PopLog();

    #
    # Compare generated LSL traits to expected traits
    #
    $testName = "Expected LSL output";
    LogMsg("$testName...\n");
    PushLog();
    foreach $trait (@traitResults){
	local($resultTrait) = "$resultDir$trait";
	local($expectTrait) = "$expectDir$trait";
	LogMsg("Compare trait $resultTrait...");
	local($status) = DiffExpected($resultTrait, $expectTrait);
	if($status == 0){
	    LogMsg("$passed\n");
	    TestResults($fileBase, $testName, $passed, [], []);
	} elsif ($status == -1){
	    $passTest = 0;
	    LogMsg("$failed\n");
	    TestResults($fileBase, $testName, $failed, 
			["Unable to locate the expected output file!"],
			[$expectTrait]);
	} else {
	    LogMsg("$failed\n");
	    $passTest = 0;
	    TestResults($fileBase, $testName, $failed, 
			["There is a difference between the expected LSL trait",
			 "and the generated LSL trait"],
			[$resultTrait, $expectTrait]);
	}
    }
    PopLog();
    
    
    
    PopLog();
    
    return $passTest;
}

sub VspecTest {
    PushLog();
    # Process arguments
    local($fileBase) = shift @_;
    local($passTest) = 0;
    
    local($testName) = "Publish VSPEC";
    LogMsg("$testName...");
    local($resultFile) = "$resultDir$fileBase.vspectest";
    local($expectFile) = "$expectDir$fileBase.vspectest";
    RemoveResultFile($resultFile);
    
    local($status) = system("$vsp_cmd -print-vspec $file 1> $resultFile 2> /dev/null");
    
    if($status == 0){
	LogMsg("$passed\n");
	TestResults($fileBase, $testName, $passed, [], []);
    } else {
	LogMsg("$failed\n");
	$passTest = 0;
	TestResults($fileBase, $testName, $failed, 
		    ["The parser returned an error when publishing VSPEC"],
		    [$resultFile]);
    }
    
    $testName = "Parse VSPEC";
    LogMsg("$testName...");
    local($status) = system("$vsp_cmd $resultFile 2> /dev/null 1>> /dev/null");
    if($status == 0){
	LogMsg("$passed\n");
	TestResults($fileBase, $testName, $passed, [], []); 
    } else {
	$passTest = 0;
	LogMsg("$failed\n");
	TestResults($fileBase, $testName, $failed, 
		    ["The parser was unable to re-parse the generated VSPEC"],
		    [$resultFile]);
    }
    
    $testName = "Expected VSPEC output";
    LogMsg("$testName...");
    $status = DiffExpected($resultFile, $expectFile);
    if($status == 0){
	LogMsg("$passed\n");
	TestResults($fileBase, $testName, $passed, [], []);
    } elsif($status == -1){
	$passTest = 0;
	LogMsg("$failed\n");
	TestResults($fileBase, $testName, $failed, 
		    ["Unable to locate the expected output file!"],
		    [$expectFile]);
    } else {
	$passTest = 0;
	LogMsg("$failed\n");
	TestResults($fileBase, $testName, $failed, 
		    ["There is a difference between the expected VSPEC file",
		     "generated by the parser, and the actual VSPEC file ",
		     "generated by the parser."],
		    [$resultFile, $expectFile]);
    }
    
    PopLog();
    
    return $passTest;
}

sub CleanTests {
    # Process arguments
    local($cleanDir) = shift @_;
    
    #
    # Get list of test files
    #
    opendir(DIR, $cleanDir);
    my(@files) = readdir(DIR);
    closedir(DIR);
    
    # Go to test director
    chdir "$cleanDir" ||
	die "Can't locate clean test directory $cleanDir";
    
    # Setup results directory
    if(! (-d "results")){
	LogMsg("Creating results directory\n");
	mkdir("results",0777) ||
	    die "Unable to create results directory";
    }
    
    # Process test files
    foreach $file (@files) {
	if($file =~ /\S+\.sld$/i){
	    LogMsg("Testing file: $file\n");
	    #
	    # Run parse tests
	    #
	    if(ParseTest($file) == 0){
		
		#
		# Run Vspec tests
		# These tests are disabled until the Savant guys
		# fix the VHDL publishing bugs.
		# VspecTest($file);
		
		#
		# Run LSL tests
		#
		# LSLTest($file);
	    } else {
		LogMsg("Skipping other tests\n");
	    }
	}
    }
}


#===================
#
# Main Program
#
#===================

#
# Global flags
#
$updateExpected = "false";

#
# Process Arguments
#
@ARGV = ('-') unless @ARGV;
while($ARGV = shift){
    if($ARGV =~ /--help/ || $ARGV =~ /-help/ || $ARGV =~ /-h/){
	print "Usage:\n";
	print "  --help     print this message\n";
	print "  -force_update   Overwrite the expected results with the\n";
	print "                  generated results\n";
	print "  -query_update   Ask the use if the expected results should be\n";
	print "                  updated when they differ from the actual results\n";
	
    } elsif($ARGV =~ /-force_update/){
	$updateExpected = "force";
    } elsif($ARGV =~ /-query_update/){
	$updateExpected = "query";
    } elsif ($ARGV =~ /-do_update/){
	$updateExpected = "do";
	}
    
}


#
# Setup environment variables
#
#  $uctoolsroot_env_var = "UCTOOLSROOT";
#  defined($uctoolsroot = $ENV{$uctoolsroot_env_var}) || 
#      die "Environment Variable UCTOOLSROOT Undefined.\n";

$curdir = cwd();
#defined($sldlhome = $ENV{$sldlhome}) || die "Environment SLDL_HOME Undefined\n";
#
# Locate executables
#

#$vsp_cmd = "/users/krishna/development/sldl/sldl-lrm/grammar/src/sldl";
$vsp_cmd = "$curdir/../src/sldl";

# $lsl_cmd = "/bin/lsl";

(-e $vsp_cmd) || 
    die "could not find sldl in $vsp_cmd.\nCheck your UCTOOLSROOT env variable\n"; 
#(-e $lsl_cmd) || 
#    die "could not find lsl in $lsl_cmd.\nCheck your UCTOOLSROOT env variable\n"; 

#
# Setup global variables
#
$logfile = STDERR;
$logLevel = 0;
$resultDir = "results/";
$expectDir = "expect/";
$passed = "PASSED";
$failed = "FAILED";

$homedir = cwd();

#
# Run clean tests
#
chdir($homedir);
if(-d "clean"){
    CleanTests("clean");
} else {
    LogMsg("Unable to locate 'clean' test directory!\n");
}

#
# Run dirty tests
#
chdir($homedir);
if(-d "dirty"){
    DirtyTests("dirty");
} else {
    LogMsg("Unable to locate 'dirty' test directory!\n");
}


#
# Output summary
#
chdir($homedir);
SummarizeResults();































